@@ -1,4 +1,4 @@
-/*  $Id: blastdbcmd.cpp 631510 2021-05-19 13:47:40Z ivanov $
+/*  $Id: blastdbcmd.cpp 674111 2023-10-18 18:11:32Z ivanov $
  * ===========================================================================
  *
  *                            PUBLIC DOMAIN NOTICE
@@ -34,6 +34,7 @@
 
 #include <ncbi_pch.hpp>
 #include <corelib/ncbiapp.hpp>
+#include <serial/objostrjson.hpp>
 #include <algo/blast/api/version.hpp>
 #include <objtools/blast/seqdb_reader/seqdbexpert.hpp>
 #include <objtools/blast/seqdb_reader/impl/seqdbtax.hpp>
@@ -42,8 +43,11 @@
 #include <objtools/blast/blastdb_format/seq_formatter.hpp>
 #include <objtools/blast/blastdb_format/blastdb_formatter.hpp>
 #include <objtools/blast/blastdb_format/blastdb_seqid.hpp>
+#include <objtools/blast/seqdb_reader/seqdbcommon.hpp>
+#include <objtools/blast/seqdb_reader/tax4blastsqlite.hpp> // for taxid to their descendant taxids lookup
 #include <algo/blast/blastinput/blast_input.hpp>
 #include <objects/seqloc/PDB_seq_id.hpp>
+#include <serial/objostrjson.hpp>
 #include "../blast/blast_app_util.hpp"
 #include <iomanip>
 
@@ -331,8 +335,26 @@ CBlastDBCmdApp::x_InitBlastDB_TaxIdList()
     	    }
     	}
     }
-    for(unsigned int i=0; i < ids.size(); i++) {
-    	m_TaxIdList.insert(NStr::StringToNumeric<TTaxId>(ids[i], NStr::fAllowLeadingSpaces | NStr::fAllowTrailingSpaces));
+
+    unique_ptr<ITaxonomy4Blast> tb;
+    if( ! args[kArgNoTaxIdExpansion].AsBoolean() ){
+        try{
+            tb.reset(new CTaxonomy4BlastSQLite());
+        }
+        catch(CException &){
+            LOG_POST(Warning << "The -taxids command line option requires additional data files. Please see the section 'Taxonomic filtering for BLAST databases' in https://www.ncbi.nlm.nih.gov/books/NBK569839/ for details.");
+        }
+    }
+
+    for (auto id : ids) {
+        auto taxid = NStr::StringToNumeric<TTaxId>(id, NStr::fAllowLeadingSpaces | NStr::fAllowTrailingSpaces);
+    	m_TaxIdList.insert(taxid);
+        if (tb) {
+            vector<int> descendants;
+            tb->GetLeafNodeTaxids(taxid, descendants);
+            for (auto d: descendants)
+                m_TaxIdList.insert(static_cast<TTaxId>(d));
+        }
     }
 
     CSeqDB::ESeqType seqtype = ParseMoleculeTypeString(args[kArgDbType].AsString());
@@ -559,25 +581,18 @@ private:
 		eCommonName,
 		eSuperKingdom,
 		eBlastName,
+        eNumSeqs,
 		eMaxFields
 	};
 	CNcbiOstream & m_Out;
 	vector<int> m_Fields;
 	vector<string> m_Seperators;
 	bool m_NeedTaxInfoLookup;
+	bool m_NeedNumSeqs;
 public:
-	CPrintTaxFields(CNcbiOstream & out, const string & fmt): m_Out(out), m_NeedTaxInfoLookup(true) {
+	CPrintTaxFields(CNcbiOstream & out, const string & fmt): m_Out(out), m_NeedTaxInfoLookup(false), m_NeedNumSeqs(false) {
 		vector<string> fields;
         string sp = kEmptyStr;
-		if(fmt == "%f") {
-			m_Seperators.push_back(sp);
-			for(unsigned int i=eTaxID; i < eMaxFields; i++){
-				m_Fields.push_back(i);
-				m_Seperators.push_back("\t");
-			}
-			return;
-		}
-
 	    for (unsigned int i = 0; i < fmt.size(); i++) {
 	    	if (fmt[i] == '%') {
 		        if (fmt[i+1] == '%') {
@@ -586,20 +601,37 @@ public:
 		        }
 		        i++;
 		        switch (fmt[i]) {
+		        case 'f' :
+		        	m_NeedTaxInfoLookup = true;
+		      		m_Seperators.push_back(sp);
+		      		for(unsigned int i=eTaxID; i < eMaxFields; i++){
+		                if ( i == eNumSeqs ) continue;
+		      		    m_Fields.push_back(i);
+		      		    m_Seperators.push_back("\t");
+		      		}
+		        break;
 		        case 'T' :
        				m_Fields.push_back(eTaxID);
        			break;
 		        case 'S' :
 		        	m_Fields.push_back(eSciName);
+			        m_NeedTaxInfoLookup = true;
                 break;
 		        case 'L' :
 		        	m_Fields.push_back(eCommonName);
+			        m_NeedTaxInfoLookup = true;
 		        break;
 		        case 'K' :
 		        	m_Fields.push_back(eSuperKingdom);
+			        m_NeedTaxInfoLookup = true;
                 break;
 		        case 'B' :
 		        	m_Fields.push_back(eBlastName);
+			        m_NeedTaxInfoLookup = true;
+		        break;
+		        case 'n' :
+		        	m_Fields.push_back(eNumSeqs);
+                    m_NeedNumSeqs = true;
 		        break;
 		        default:
 	                sp += fmt[i-1];
@@ -620,12 +652,9 @@ public:
 			NCBI_THROW(CInputException, eInvalidInput,
 				       "Invalid format options for tax_info.");
 		}
-		if((m_Fields.size() == 1) && (m_Fields[0] == eTaxID)){
-			m_NeedTaxInfoLookup = false;
-		}
 	}
 
-	void PrintEntry(const SSeqDBTaxInfo & t){
+	void PrintEntry(const SSeqDBTaxInfo & t, int num_seqs){
 		for(unsigned int i=0; i < m_Fields.size(); i++) {
 			m_Out << m_Seperators[i];
 			switch (m_Fields[i]){
@@ -644,6 +673,9 @@ public:
 				case eBlastName:
 					m_Out << t.blast_name;
 				break;
+				case eNumSeqs:
+					m_Out << num_seqs;
+				break;
 				default:
 					NCBI_THROW(CInputException, eInvalidInput,
 					           "Invalid format options for tax_info.");
@@ -654,6 +686,7 @@ public:
 		m_Out << "\n";
 	}
 	bool NeedTaxNames(){return m_NeedTaxInfoLookup;}
+	bool NeedNumSeqs(){return m_NeedNumSeqs;}
 };
 
 
@@ -668,6 +701,7 @@ CBlastDBCmdApp::x_PrintBlastDatabaseTaxInformation()
     CPrintTaxFields tf(out, kFmt);
     set<TTaxId> tax_ids;
     m_BlastDb->GetDBTaxIds(tax_ids);
+
     // Print basic database information
     out << "# of Tax IDs in Database: " << tax_ids.size() << endl;
 	SSeqDBTaxInfo info;
@@ -686,7 +720,15 @@ CBlastDBCmdApp::x_PrintBlastDatabaseTaxInformation()
     	else {
    			info.taxid = *itr;
     	}
-   		tf.PrintEntry(info);
+        int num_seqs = 0;
+        if(tf.NeedNumSeqs()) {
+            vector<blastdb::TOid> rv;
+            set<TTaxId> ti;
+            ti.insert(*itr);
+            m_BlastDb->TaxIdsToOids(ti, rv);
+            num_seqs = rv.size();
+        }
+   		tf.PrintEntry(info, num_seqs);
     }
 }
 
@@ -860,7 +902,7 @@ void CBlastDBCmdApp::Init()
 {
     HideStdArgs(fHideConffile | fHideFullVersion | fHideXmlHelp | fHideDryRun);
 
-    auto_ptr<CArgDescriptions> arg_desc(new CArgDescriptions);
+    unique_ptr<CArgDescriptions> arg_desc(new CArgDescriptions);
 
     // Specify USAGE context
     arg_desc->SetUsageContext(GetArguments().GetProgramBasename(),
@@ -910,30 +952,34 @@ void CBlastDBCmdApp::Init()
         arg_desc->SetDependency("ipg_batch", CArgDescriptions::eExcludes, "strand");
         arg_desc->SetDependency("ipg_batch", CArgDescriptions::eExcludes, "mask_sequence_with");
 
-    arg_desc->AddOptionalKey(kArgTaxIdList, "taxonomy_ids",
-    						"Comma-delimited taxonomy identifiers", CArgDescriptions::eString);
-    arg_desc->SetDependency(kArgTaxIdList, CArgDescriptions::eExcludes, "entry");
-    arg_desc->SetDependency(kArgTaxIdList, CArgDescriptions::eExcludes, "entry_batch");
-    arg_desc->SetDependency(kArgTaxIdList, CArgDescriptions::eExcludes, "pig");
-
-    arg_desc->AddOptionalKey(kArgTaxIdListFile, "input_file",
-    						"Input file for taxonomy identifiers", CArgDescriptions::eInputFile);
-    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, "entry");
-    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, "entry_batch");
-    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, "pig");
-    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, kArgTaxIdList);
-
     arg_desc->AddFlag("info", "Print BLAST database information", true);
     // All other options to this program should be here
     const char* exclusions[]  = { "entry", "entry_batch", "outfmt", "strand",
         "target_only", "ctrl_a", "get_dups", "pig", "range",
         "mask_sequence", "list", "remove_redundant_dbs", "recursive",
-        "list_outfmt", kArgTaxIdListFile.c_str(), kArgTaxIdList.c_str()};
+        "list_outfmt", "metadata", "metadata_output_prefix", kArgTaxIdListFile.c_str(), kArgTaxIdList.c_str(),
+        kArgNoTaxIdExpansion.c_str()};
     for (size_t i = 0; i < sizeof(exclusions)/sizeof(*exclusions); i++) {
         arg_desc->SetDependency("info", CArgDescriptions::eExcludes,
                                 string(exclusions[i]));
     }
 
+    arg_desc->AddFlag("metadata", "Generate BLAST database metadata", true);
+    // All other options to this program should be here
+    const char* exclusions_m[]  = { "entry", "entry_batch", "outfmt", "strand",
+        "target_only", "ctrl_a", "get_dups", "pig", "range",
+        "mask_sequence", "list", "remove_redundant_dbs", "recursive",
+        "list_outfmt", "info", kArgTaxIdListFile.c_str(), kArgTaxIdList.c_str(),
+        kArgNoTaxIdExpansion.c_str()};
+    for (size_t i = 0; i < sizeof(exclusions_m)/sizeof(*exclusions_m); i++) {
+        arg_desc->SetDependency("metadata", CArgDescriptions::eExcludes,
+                                string(exclusions_m[i]));
+    }
+
+    arg_desc->AddOptionalKey("metadata_output_prefix", "",
+    						"Path prefix for location of database files in metadata", CArgDescriptions::eString);
+    arg_desc->SetDependency("metadata_output_prefix", CArgDescriptions::eRequires, "metadata");
+
     arg_desc->AddFlag("tax_info",
     		          "Print taxonomic information contained in this BLAST database.\n"
     		          "Use -outfmt to customize output. Format specifiers supported are:\n"
@@ -942,12 +988,14 @@ void CBlastDBCmdApp::Init()
     		          "\t\t%S means scientific name\n"
     		          "\t\t%K means taxonomic super kingdom\n"
     		          "\t\t%B means BLAST name\n"
+    		          "\t\t%n means num of seqs\n"
     		          "By default it prints: '%T %S %L %K %B'\n", true);
     // All other options to this program should be here
     const char* tax_info_exclusions[]  = { "info", "entry", "entry_batch", "strand",
         "target_only", "ctrl_a", "get_dups", "pig", "range",
         "mask_sequence", "list", "remove_redundant_dbs", "recursive",
-        "list_outfmt", kArgTaxIdListFile.c_str(), kArgTaxIdList.c_str() };
+        "list_outfmt", kArgTaxIdListFile.c_str(), kArgTaxIdList.c_str(),
+        kArgNoTaxIdExpansion.c_str() };
     for (size_t i = 0; i < sizeof(tax_info_exclusions)/sizeof(*tax_info_exclusions); i++) {
         arg_desc->SetDependency("tax_info", CArgDescriptions::eExcludes,
                                 string(tax_info_exclusions[i]));
@@ -970,6 +1018,26 @@ void CBlastDBCmdApp::Init()
                              "algorithm ID specified",
                              CArgDescriptions::eString);
 
+    // Some additional tax related grouping
+    arg_desc->SetCurrentGroup("Taxonomic filtering options"); 
+    arg_desc->AddOptionalKey(kArgTaxIdList, "taxonomy_ids",
+    						"Comma-delimited taxonomy identifiers", CArgDescriptions::eString);
+    arg_desc->SetDependency(kArgTaxIdList, CArgDescriptions::eExcludes, "entry");
+    arg_desc->SetDependency(kArgTaxIdList, CArgDescriptions::eExcludes, "entry_batch");
+    arg_desc->SetDependency(kArgTaxIdList, CArgDescriptions::eExcludes, "pig");
+
+    arg_desc->AddOptionalKey(kArgTaxIdListFile, "input_file",
+    						"Input file for taxonomy identifiers", CArgDescriptions::eInputFile);
+    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, "entry");
+    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, "entry_batch");
+    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, "pig");
+    arg_desc->SetDependency(kArgTaxIdListFile, CArgDescriptions::eExcludes, kArgTaxIdList);
+
+    // Disable Tax ID resoution to the descendants SB-3791/SB-3779 
+    arg_desc->AddFlag(kArgNoTaxIdExpansion, "Do not expand the taxonomy IDs provided to their descendant taxonomy IDs ", true);
+    arg_desc->SetDependency(kArgNoTaxIdExpansion, CArgDescriptions::eExcludes, "entry");
+    arg_desc->SetDependency(kArgNoTaxIdExpansion, CArgDescriptions::eExcludes, "entry_batch");
+    arg_desc->SetDependency(kArgNoTaxIdExpansion, CArgDescriptions::eExcludes, "pig");
     arg_desc->SetCurrentGroup("Output configuration options");
     arg_desc->AddDefaultKey(kArgOutput, "output_file", "Output file name",
                             CArgDescriptions::eOutputFile, "-");
@@ -1135,6 +1203,23 @@ int CBlastDBCmdApp::Run(void)
         	x_InitBlastDB();
             x_PrintBlastDatabaseInformation();
         }
+        else if (args["metadata"]) {
+        	x_InitBlastDB();
+            string output_prefix = args["metadata_output_prefix"]
+                    ? args["metadata_output_prefix"].AsString()
+                    : kEmptyStr;
+            if (!output_prefix.empty() && (output_prefix.back() != CFile::GetPathSeparator()))
+                output_prefix += CFile::GetPathSeparator();
+        	CRef<CBlast_db_metadata> m = m_BlastDb->GetDBMetaData(output_prefix);
+            unique_ptr<CObjectOStreamJson> json_out(new CObjectOStreamJson(out, eNoOwnership));
+            json_out->SetDefaultStringEncoding(eEncoding_Ascii);
+            json_out->PreserveKeyNames();
+            CConstObjectInfo obj_info(m, m->GetTypeInfo());
+            json_out->WriteObject(obj_info);
+            json_out->Flush();
+            out.flush();
+            out << NcbiEndl;
+        }
         else if (args["tax_info"]) {
         	x_InitBlastDB();
             x_PrintBlastDatabaseTaxInformation();
@@ -1184,8 +1269,16 @@ void CBlastDBCmdApp::x_AddCmdOptions()
     	m_UsageReport.AddParam(CBlastUsageReport::eOutputFmt, args["outfmt"].AsString());
     }
 
+	vector<string> db_list;
+	NStr::Split(m_BlastDb->GetDBNameList(), " ", db_list,NStr::fSplit_Tokenize);
+	NON_CONST_ITERATE(vector<string>, itr, db_list) {
+		int off = (*itr).find_last_of(CFile::GetPathSeparator());
+	    if (off != -1) {
+	        (*itr).erase(0, off+1);
+	    }
+	}
+	string db_name = NStr::Join(db_list, " ");
 
-	string db_name = m_BlastDb->GetDBNameList();
 	int off = db_name.find_last_of(CFile::GetPathSeparator());
     if (off != -1) {
     	db_name.erase(0, off+1);
